# ------------------------------
# Import libraries
# ------------------------------

import zipfile
import yaml
from collections import OrderedDict
from typing import Dict, List, Optional, Tuple, Union
import os
import sys
import glob
import shutil
from copy import deepcopy
import numpy as np
import random
import torch
import torch.nn as nn
from torch.optim import lr_scheduler
from torch.utils.data import Dataset
import torchvision
from torchvision.models import efficientnet_v2_s, vgg19, resnet50, inception_v3, densenet121
from torchvision.utils import make_grid, save_image
import cv2
# import albumentations as A
import flwr as fl
from flwr.common import (
    Code,
    EvaluateIns,
    EvaluateRes,
    FitIns,
    FitRes,
    GetParametersIns,
    GetParametersRes,
    Status,
    NDArrays,
    Parameters,
    Scalar,
    ndarrays_to_parameters,
    parameters_to_ndarrays,
)
from flwr.server.strategy import FedAvg
from strategy import PlainFedAvg



# ------------------------------
# Load config file
# ------------------------------

config_file_name = "config.yaml"
with open(config_file_name) as file:
    CONFIG = yaml.safe_load(file)



# ------------------------------
# Extract .zip folder
# ------------------------------

def extract_zip(zip_file_name: str, save_dir: str):
    """
    Extract .zip folder, if not already done
    """

    parts = zip_file_name.split(".")
    file_name = parts[0]

    if not os.path.isdir(file_name):
        with zipfile.ZipFile(zip_file_name, "r") as zip_ref:
            zip_ref.extractall(save_dir)
        print(f"Folder '{zip_file_name}' was successfully extracted")
    else:
        print(f"Folder '{file_name}' already exists")



# ------------------------------
# Set seeds
# ------------------------------

def set_seeds(seed):
    """

    """

    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False



# ----------------------------------------
#
# ----------------------------------------

class_map_str_to_idx = {
    "No_windshield": 0,
    "Windshield_TypeA": 1,
    "Windshield_TypeB": 2,
    "Windshield_TypeC": 3,
    "Windshield_TypeD": 4
}



# ----------------------------------------
# Load data
# ----------------------------------------

class Dataloader(Dataset):
    def __init__(self, root_dir, loader_mode, imgsz):
        self.root_dir = root_dir
        self.loader_mode = loader_mode
        if self.loader_mode == "test":
            self.imgs = [elem for elem in glob.glob(os.path.join(root_dir, "*", "*"))]
        else:
            self.imgs = [elem for elem in glob.glob(os.path.join(root_dir, self.loader_mode, "*", "*"))]

        self.imgs_and_lbls = []
        for path in self.imgs:
            self.imgs_and_lbls.append([cv2.imread(path, cv2.IMREAD_COLOR), class_map_str_to_idx[path.split(os.sep)[-2]]])

        self.transform = torchvision.transforms.Compose([
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Resize((imgsz, imgsz)),
        ])

        # if self.loader_mode == "train":
        #     self.augment = A.Compose([
        #         A.VerticalFlip(p=0.5),
        #         A.Rotate(limit=15, border_mode=cv2.BORDER_CONSTANT, value=(0,0,0), p=0.75)
        #     ])

    def __len__(self):
        return len(self.imgs)

    def __getitem__(self, idx):
        img, lbl = self.imgs_and_lbls[idx]

        # add augmentation during training
        # if self.loader_mode == "train":
        #     augmented = self.augment(img)
        #     img = augmented["image"]

        # transform numpy array into Pytorch tensor
        img_torch = self.transform(img)

        return img_torch, lbl



def load_train_val_data(root_dir, batch_size, imgsz):
    """

    """

    train_dataset = Dataloader(root_dir=root_dir, loader_mode="train", imgsz=imgsz)
    val_dataset = Dataloader(root_dir=root_dir, loader_mode="val", imgsz=imgsz)

    train_dataloader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, pin_memory=True, shuffle=True)
    val_dataloader = torch.utils.data.DataLoader(val_dataset, batch_size=batch_size, pin_memory=True, shuffle=False)

    return train_dataloader, val_dataloader



def load_test_data(root_dir, batch_size, imgsz):
    """

    """

    test_dataset = Dataloader(root_dir=root_dir, loader_mode="test", imgsz=imgsz)

    test_dataloader = torch.utils.data.DataLoader(test_dataset, batch_size=batch_size, pin_memory=True, shuffle=False)

    return test_dataloader



# ----------------------------------------
# Train model
# ----------------------------------------

def train_model(model,
                train_dataloader,
                val_dataloader,
                optimizer,
                num_epochs,
                criterion,
                device,
                use_best_weights,
                early_stopping,
                save_dir=None):
    """

    """

    # if CONFIG["lr_scheduler"]:
    #     scheduler = lr_scheduler.StepLR(optimizer, step_size=3, gamma=0.75)

    best_loss = np.inf

    model.to(device)
    model.train()

    hist_train_loss, hist_train_acc, hist_val_loss, hist_val_acc = [], [], [], []

    epochs_without_improvement = 0

    for epoch in range(num_epochs):
        # training
        model.train()

        correct = 0
        total = 0
        epoch_train_loss = 0.0

        for inputs, labels in train_dataloader:
            inputs = inputs.to(device)
            labels = labels.to(device)

            optimizer.zero_grad()

            outputs = model(inputs)
            _, preds = torch.max(nn.functional.softmax(outputs, dim=1), 1)
            loss = criterion(outputs, labels)

            loss.backward()
            optimizer.step()

            epoch_train_loss += loss.item()
            correct += torch.sum(preds == labels).item()
            total += labels.size(0)

        epoch_train_loss = epoch_train_loss / total
        epoch_train_acc = correct / total
        hist_train_loss.append(epoch_train_loss)
        hist_train_acc.append(epoch_train_acc)

        # scheduler.step()

        # validation
        model.eval()

        correct = 0
        total = 0
        epoch_val_loss = 0.0

        with torch.no_grad():
            for inputs, labels in val_dataloader:
                inputs = inputs.to(device)
                labels = labels.to(device)

                outputs = model(inputs)
                _, preds = torch.max(nn.functional.softmax(outputs, dim=1), 1)

                epoch_val_loss += criterion(outputs, labels).item()
                correct += torch.sum(preds == labels).item()
                total += labels.size(0)

        epoch_val_loss = epoch_val_loss / total
        epoch_val_acc = correct / total
        hist_val_loss.append(epoch_val_loss)
        hist_val_acc.append(epoch_val_acc)

        # early stopping
        if epoch_val_loss < best_loss:
            best_loss = epoch_val_loss
            epochs_without_improvement = 0

            # save best model as .pt
            save_model(model, os.path.join(save_dir, "best.pt"))
            if use_best_weights:
                best_model_state = deepcopy(model.state_dict())
        else:
            epochs_without_improvement += 1
            if epochs_without_improvement >= early_stopping:
                print(f"Early stopping after {epoch+1} epochs without improvement")
                break

    history = {"train_acc": hist_train_acc, "train_loss": hist_train_loss, "val_acc": hist_val_acc, "val_loss": hist_val_loss}

    # save last model as .pt
    save_model(model, os.path.join(save_dir, "last.pt"))

    if use_best_weights:
        model.load_state_dict(best_model_state)

    return history



# ----------------------------------------
# Test model
# ----------------------------------------

def test_model(model, dataloader, criterion, device, num_classes):
    """
    Test model and calculate confusion matrix
    """

    model.to(device)
    model.eval()

    confusion_matrix = np.zeros((num_classes, num_classes))

    correct = 0
    total = 0
    loss = 0.0

    with torch.no_grad():
        for inputs, labels in dataloader:
            inputs = inputs.to(device)
            labels = labels.to(device)

            outputs = model(inputs)
            _, preds = torch.max(nn.functional.softmax(outputs, dim=1), 1)

            loss += criterion(outputs, labels).item()
            correct += torch.sum(preds == labels).item()
            total += labels.size(0)

            pred_cls = preds.cpu().numpy()
            real_cls = labels.cpu().numpy()
            for pred_cls_i, real_cls_i in zip(pred_cls, real_cls):
                confusion_matrix[int(pred_cls_i), int(real_cls_i)] += 1

    loss = loss / total
    acc = correct / total

    return acc, loss, confusion_matrix



# ----------------------------------------
# Calculate metrics
# ----------------------------------------

def calculate_metrics(confusion_matrix):
    """

    """

    num_classes = len(confusion_matrix)

    Recall_list = []
    Precision_list = []
    F1_list = []

    for i in range(num_classes):
        TP_cls_i = confusion_matrix[i,i]
        FP_cls_i = sum(confusion_matrix[i,:]) - TP_cls_i
        FN_cls_i = sum(confusion_matrix[:,i]) - TP_cls_i

        Recall_i = TP_cls_i/(TP_cls_i + FN_cls_i)
        Precision_i = TP_cls_i/(TP_cls_i + FP_cls_i)
        F1_i = 2*(Precision_i * Recall_i)/(Precision_i + Recall_i)

        if np.isnan(Recall_i):
            Recall_i = 0
        if np.isnan(Precision_i):
            Precision_i = 0
        if np.isnan(F1_i):
            F1_i = 0

        Recall_list.append(Recall_i)
        Precision_list.append(Precision_i)
        F1_list.append(F1_i)

    average_Recall = np.array(Recall_list).mean()
    average_Precision = np.array(Precision_list).mean()
    average_F1 = np.array(F1_list).mean()

    return average_Precision, average_Recall, average_F1



# ----------------------------------------
# Save metrics
# ----------------------------------------

def save_metrics(metrics: Dict[str, Scalar], header, filename: str):
    """
    Write results to a text file
    """

    keys, vals = list(metrics.keys()), np.array(list(metrics.values())).T
    num_cols = len(metrics) + 1
    num_rows = vals.shape[0]
    header = (("%23s," * num_cols % tuple([header] + keys)).rstrip(","))

    with open(filename, "w") as f:
        f.write(header + "\n")
        for index in range(num_rows):
            f.write(("%23.5g," * num_cols % tuple([index+1] + list(vals[index]))).rstrip(",") + "\n")

    return



# ------------------------------
# Get model
# ------------------------------

def get_model(model_config: str, num_classes: int):
    """

    """

    if model_config == "EfficientNet":
        model = efficientnet_v2_s(weights="IMAGENET1K_V1")
        in_features = model.classifier[-1].in_features
        model.classifier[-1] = torch.nn.Linear(in_features, num_classes)
    elif model_config == "VGG":
        model = vgg19(weights="IMAGENET1K_V1")
        in_features = model.classifier[-1].in_features
        model.classifier[-1] = torch.nn.Linear(in_features, num_classes)
    elif model_config == "ResNet":
        model = resnet50(weights="IMAGENET1K_V1")
        model.fc = torch.nn.Linear(model.fc.in_features, num_classes)
    elif model_config == "InceptionNet":
        model = inception_v3(weights="IMAGENET1K_V1")
        model.aux_logits = False
        model.fc = torch.nn.Linear(model.fc.in_features, num_classes)
    elif model_config == "DenseNet":
        model = densenet121(weights="IMAGENET1K_V1")
        model.classifier = torch.nn.Linear(model.classifier.in_features, num_classes)

    return model



# ------------------------------
# Get strategy
# ------------------------------

def get_strategy(strategy_config: str):
    """

    """

    if strategy_config == "FedAvg":
        strategy = FedAvg(fraction_fit=1.0,
                          fraction_evaluate=0.0,
                          min_fit_clients=CONFIG["num_clients"],
                          min_evaluate_clients=CONFIG["num_clients"],
                          min_available_clients=CONFIG["num_clients"],
                          on_fit_config_fn=fit_config,
                          evaluate_fn=evaluate)
    elif strategy_config == "PlainFedAvg":
        strategy = PlainFedAvg(fraction_fit=1.0,
                               fraction_evaluate=0.0,
                               min_fit_clients=CONFIG["num_clients"],
                               min_evaluate_clients=CONFIG["num_clients"],
                               min_available_clients=CONFIG["num_clients"],
                               on_fit_config_fn=fit_config,
                               evaluate_fn=evaluate)

    return strategy



# ------------------------------
# Get optimizer
# ------------------------------

def get_optimizer(model, optimizer_config: str):
    """

    """

    if optimizer_config == "SGD":
        optimizer = torch.optim.SGD(model.parameters(),
                                    lr=CONFIG["lr"],
                                    momentum=CONFIG["momentum"])
    elif optimizer_config == "Adam":
        optimizer = torch.optim.Adam(model.parameters(),
                                     lr=CONFIG["lr"],
                                     momentum=CONFIG["momentum"])
    elif optimizer_config == "AdamW":
        optimizer = torch.optim.AdamW(model.parameters(),
                                     lr=CONFIG["lr"],
                                     momentum=CONFIG["momentum"])

    return optimizer



# ------------------------------
# Get criterion
# ------------------------------

def get_criterion(criterion_config: str):
    """

    """

    if criterion_config == "CrossEntropyLoss":
        criterion = nn.CrossEntropyLoss()

    return criterion



# ------------------------------
# Create folder
# ------------------------------

def create_folder(dir: str):
    """
    Check if folder already exists
    If not create new folder
    """

    if not os.path.exists(dir):
        os.makedirs(dir)



# ------------------------------
# Save model as .pt
# ------------------------------

def save_model(model, save_dir):
    """
    Export model to TorchScript and save it
    """

    if not os.path.exists(save_dir):
        model_scripted = torch.jit.script(model)
        model_scripted.save(save_dir)



# ------------------------------
# Get parameters
# ------------------------------

def get_parameters(model) -> NDArrays:
    """
    Return model parameters as a list of NumPy ndarrays
    """

    return [val.cpu().numpy() for _, val in model.state_dict().items()]



# ------------------------------
# Set parameters
# ------------------------------

def set_parameters(model, parameters: NDArrays) -> None:
    """
    Set model parameters from a list of NumPy ndarrays
    """

    params_dict = zip(model.state_dict().keys(), parameters)
    state_dict = OrderedDict({k: torch.tensor(v) for k, v in params_dict})
    model.load_state_dict(state_dict, strict=True)

    return



# ------------------------------
# Implement Flower client
# ------------------------------

class FlowerClient(fl.client.Client):
    """
    Flower client using PyTorch
    """

    def __init__(self, cid, model, train_dataloader, val_dataloader, device, root_dir, save_dir):
        self.cid = cid
        self.model = model
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader
        self.device = device
        self.root_dir = root_dir
        self.save_dir = save_dir


    def get_parameters(self,
                       ins: GetParametersIns,
                       ) -> GetParametersRes:
        """
        Return model parameters as a list of NumPy ndarrays
        """

        # serialize ndarrays parameters into Parameters object
        parameters = ndarrays_to_parameters(get_parameters(self.model))

        # build and return response
        return GetParametersRes(
            status=Status(code=Code.OK, message="Success"),
            parameters=parameters,
        )


    def fit(self,
            ins: FitIns,
            ) -> FitRes:
        """
        Set model parameters, train model, return updated model parameters
        """

        # create folder
        dir = os.path.join(self.save_dir, f"server_round_{ins.config['server_round']}")
        create_folder(dir)

        # deserialize parameters to Numpy ndarrays
        parameters_ndarrays = parameters_to_ndarrays(ins.parameters)

        # update local model
        set_parameters(self.model, parameters_ndarrays)

        # specify optimizer
        optimizer = get_optimizer(self.model, ins.config["optimizer"])

        # specify loss function
        criterion = get_criterion(ins.config["loss_fcn"])

        # train local model
        metrics_dict = train_model(self.model,
                                   self.train_dataloader,
                                   self.val_dataloader,
                                   optimizer,
                                   ins.config["epochs"],
                                   criterion,
                                   self.device,
                                   ins.config["use_best_weights"],
                                   ins.config["early_stopping"],
                                   dir)

        # get updated parameters and serialize ndarrays parameters into Parameters object
        parameters_updated = ndarrays_to_parameters(get_parameters(self.model))

        # write training and validation metrics to .txt file
        save_metrics(metrics_dict, "epoch", os.path.join(dir, "results.txt"))

        # build and return response
        return FitRes(
            status=Status(code=Code.OK, message="Success"),
            parameters=parameters_updated,
            num_examples=len(self.train_dataloader),
            metrics={},
        )


    def evaluate(self,
                 ins: EvaluateIns,
                 ) -> EvaluateRes:
        """
        Set model parameters, evaluate aggregated model on local test dataset, return result
        """

        # build and return response
        return EvaluateRes(
            status=Status(code=Code.OK, message="Success"),
            loss=0.0,
            num_examples=0,
            metrics={},
        )



# ------------------------------
# Evaluation function
# ------------------------------

def evaluate(server_round: int,
             parameters: NDArrays,
             config: Dict[str, Scalar],
             ) -> Optional[Tuple[float, Dict[str, Scalar]]]:
    """
    The evaluate function will be called by Flower after each server round
    """

    # create folder
    save_dir = os.path.join("runs", CONFIG["output_folder"], "global_model")
    create_folder(save_dir)

    # load pretrained model
    model = get_model(CONFIG["model"], CONFIG["num_classes"])

    # load data
    test_dataloader = load_test_data(
        os.path.join("dataset", "Cabins5_flta", "global_test_set"),
        CONFIG["batch_size"],
        CONFIG["img_size"],
    )

    # update model
    set_parameters(model, parameters)

    # save aggregated model
    save_model(model, os.path.join(save_dir, f"best_server_round_{server_round}.pt"))

    # specify loss function
    criterion = get_criterion(CONFIG["loss_fcn"])

    # test aggregated model on test dataset
    test_acc, test_loss, confusion_matrix = test_model(
        model,
        test_dataloader,
        criterion,
        DEVICE,
        CONFIG["num_classes"],
    )

    # calculate metrics
    average_Precision, average_Recall, average_F1 = calculate_metrics(confusion_matrix)

    return 0.0, {"test_loss": test_loss, "test_acc": test_acc, "precision": average_Precision, "recall": average_Recall, "f1": average_F1}



# ------------------------------
# Create Flower client
# ------------------------------

def client_fn(cid: str) -> FlowerClient:
    """
    Create a Flower client instance. Each client will
    be trained and validated on his own unique data
    """

    # load pretrained model
    model = get_model(CONFIG["model"], CONFIG["num_classes"])

    # specify output folder
    root_dir = os.path.join("runs", CONFIG["output_folder"])
    save_dir = os.path.join(root_dir, f"client_{int(cid)+1}")

    # create folder
    create_folder(save_dir)

    # load data
    train_dataloader, val_dataloader = load_train_val_data(
        os.path.join("dataset", "Cabins5_flta", f"Client{int(cid)+1}"),
        CONFIG["batch_size"],
        CONFIG["img_size"],
    )

    # save one batch of training data
    imgs, _ = next(iter(train_dataloader))
    train_batch0 = make_grid(imgs, nrow=4, scale_each=True)
    save_image(train_batch0, os.path.join(save_dir, "train_batch0"), "png")

    return FlowerClient(cid, model, train_dataloader, val_dataloader, DEVICE, root_dir, save_dir)



# ------------------------------
# Fit config function
# ------------------------------

def fit_config(server_round: int) -> Dict[str, Scalar]:
    """
    Return training configuration dict for each round
    """

    config = {
        "server_round": server_round,
        "epochs": CONFIG["epochs"],
        "optimizer": CONFIG["optimizer"],
        "momentum": CONFIG["momentum"],
        "lr": CONFIG["lr"],
        "lr_scheduler": CONFIG["lr_scheduler"],
        "loss_fcn": CONFIG["loss_fcn"],
        "use_best_weights": CONFIG["use_best_weights"],
        "early_stopping": CONFIG["early_stopping"],
    }

    return config



# ------------------------------
# Extract metrics
# ------------------------------

def extract_metrics(metrics: Dict[str, Scalar]) -> Dict[str, Scalar]:
    """

    """

    extracted_metrics = {
        "test_loss": [],
        "test_acc": [],
        "precision": [],
        "recall": [],
        "f1": [],
    }

    for key, values in metrics.items():
        for i, (_, data) in enumerate(values):
            extracted_metrics[key].append(data)

    return extracted_metrics



# ------------------------------
# Start simulation
# ------------------------------

if __name__=="__main__":
    # set seeds
    set_seeds(CONFIG["seed"])

    # define parameters
    DEVICE = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # specify output folder
    save_dir = os.path.join("runs", CONFIG["output_folder"])

    # check if folder already exists
    if os.path.exists(save_dir):
        print("Output folder already exists. Exiting the program.")
        sys.exit()

    # create output folder
    create_folder(save_dir)

    # move copy of config file to output folder
    shutil.copy(config_file_name, os.path.join(save_dir, config_file_name))

    # get strategy
    strategy = get_strategy(CONFIG["strategy"])

    # specify client resources
    client_resources = {"num_cpus": 16}
    if DEVICE.type == "cuda":
        client_resources = {"num_cpus": 16, "num_gpus": 1.0}

    # start simulation
    history = fl.simulation.start_simulation(
        client_fn=client_fn,
        num_clients=CONFIG["num_clients"],
        config=fl.server.ServerConfig(num_rounds=CONFIG["num_rounds"]),
        strategy=strategy,
        client_resources=client_resources,
    )

    # write training and validation metrics to .txt file
    extracted_metrics = extract_metrics(history.metrics_centralized)
    save_metrics(extracted_metrics, "server_round", os.path.join(save_dir, "global_model", "results_aggregated_model.txt"))
