# ------------------------------
# Import libraries
# ------------------------------

from typing import Dict, List, Optional, Tuple, Union
import numpy as np
import flwr as fl
from flwr.common import (
    FitRes,
    NDArrays,
    Parameters,
    Scalar,
    ndarrays_to_parameters,
    parameters_to_ndarrays,
)
from logging import WARNING
from flwr.common.logger import log
from flwr.server.client_proxy import ClientProxy
from functools import reduce



# ------------------------------
# Compute plain average
# ------------------------------

def aggregate_plain_average(results: List[Tuple[NDArrays, int]]) -> NDArrays:
    """
    Compute plain average
    """

    weights_list = [weights for weights, _ in results]

    # Compute average weights of each layer
    weights_prime: NDArrays = [
        reduce(np.add, layer_updates) / len(weights_list)
        for layer_updates in zip(*weights_list)
    ]

    return weights_prime



# ------------------------------
# Overwrite FedAvg class
# ------------------------------

class PlainFedAvg(fl.server.strategy.FedAvg):

    def aggregate_fit(self,
                      server_round: int,
                      results: List[Tuple[ClientProxy, FitRes]],
                      failures: List[Union[Tuple[ClientProxy, FitRes], BaseException]],
                      ) -> Tuple[Optional[Parameters], Dict[str, Scalar]]:
        """
        Aggregate fit results using plain average
        """

        if not results:
            return None, {}
        
        # do not aggregate if there are failures and failures are not accepted
        if not self.accept_failures and failures:
            return None, {}

        # Convert results
        weights_results = [
            (parameters_to_ndarrays(fit_res.parameters), fit_res.num_examples)
            for _, fit_res in results
        ]
        aggregated_ndarrays = aggregate_plain_average(weights_results)

        parameters_aggregated = ndarrays_to_parameters(aggregated_ndarrays)

        # Aggregate custom metrics if aggregation fn was provided
        metrics_aggregated = {}
        if self.fit_metrics_aggregation_fn:
            fit_metrics = [(res.num_examples, res.metrics) for _, res in results]
            metrics_aggregated = self.fit_metrics_aggregation_fn(fit_metrics)
        elif server_round == 1:  # Only log this warning once
            log(WARNING, "No fit_metrics_aggregation_fn provided")

        return parameters_aggregated, metrics_aggregated