import torch
import torch.nn as nn
import torchvision.transforms as transforms
from torchvision.models import efficientnet_v2_s, vgg19, resnet50, inception_v3, densenet121
from torchvision.datasets import ImageFolder
from torch.utils.data import DataLoader
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, confusion_matrix
import os
import matplotlib.pyplot as plt
import numpy as np
import yaml
from tqdm import tqdm
import shutil
import torch.nn.functional as F
import seaborn as sns
from PIL import Image, ImageDraw, ImageFont

# Function to read configuration from a YAML config file
def read_config(config_file):
    with open(config_file, 'r') as f:
        config = yaml.safe_load(f)
    return config

# Function to set up model architecture with configurable last layers
def setup_model(num_classes, model_name):
    model_classes = {
        "EfficientNet": efficientnet_v2_s,
        "VGG": vgg19,
        "ResNet": resnet50,
        "InceptionNet": inception_v3,
        "DenseNet": densenet121
    }
    model = model_classes[model_name](pretrained=False)

    if model_name in ["VGG", "EfficientNet"]:
        in_features = model.classifier[-1].in_features
        model.classifier[-1] = nn.Linear(in_features, num_classes)
    elif model_name == "ResNet":
        model.fc = nn.Linear(model.fc.in_features, num_classes)
    elif model_name == "DenseNet":
        model.classifier = nn.Linear(model.classifier.in_features, num_classes)
    elif model_name == "InceptionNet":
        model.aux_logits = False
        in_features = model.fc.in_features
        model.fc = nn.Linear(in_features, num_classes)

    return model

# Function to load TorchScript weights into the model
def load_weights(model, weights_path):
    loaded_model = torch.jit.load(weights_path) # Load the TorchScript model
    model.load_state_dict(loaded_model.state_dict()) # Copy the parameters from the loaded model to the provided model
    model.eval() # Set the model to evaluation mode

class_map_str_to_idx = {
    "No_windshield": 0,
    "With_windshield": 1
}

# Function to test the model
def test_model(model, test_loader):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    criterion = nn.CrossEntropyLoss()

    all_predictions = []
    all_targets = []
    all_confidences = []  # Store confidences

    with torch.no_grad(), tqdm(total=len(test_loader), desc="Running tests") as pbar:
        for images, labels in test_loader:
            images, labels = images.to(device), labels.to(device)
            outputs = model(images)
            loss = criterion(outputs, labels)

            probabilities = F.softmax(outputs, dim=1)  # Convert outputs to probabilities
            confidences, predicted = torch.max(probabilities, 1)  # Get confidences and predictions
            all_confidences.extend(confidences.cpu().numpy())  # Store confidences
            all_predictions.extend(predicted.cpu().numpy())
            all_targets.extend(labels.cpu().numpy())

            pbar.update(1)

    return all_predictions, all_targets, loss.item(), all_confidences

# Function to save results
def save_results(output_folder, all_predictions, all_targets, metrics, config, test_dataset, all_confidences):
    # Create output folder if it doesn't exist
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Create folders based on class mapping
    for class_name in class_map_str_to_idx.keys():
        class_folder = os.path.join(output_folder, class_name)
        os.makedirs(class_folder, exist_ok=True)

    # Move images to respective class folders based on predictions
    for image_path, prediction, target, confidence in zip(test_dataset.imgs, all_predictions, all_targets, all_confidences):
        image_filename = os.path.basename(image_path[0])
        try:
            predicted_class_name = [name for name, idx in class_map_str_to_idx.items() if idx == prediction][0]
            predicted_class_folder = os.path.join(output_folder, predicted_class_name)
            image_dest_path = os.path.join(predicted_class_folder, image_filename)

            # Open the image
            image = Image.open(image_path[0])

            # Write predicted class name and confidence on the top of the image
            draw = ImageDraw.Draw(image)
            font = font = ImageFont.truetype("arial.ttf", 25)  # Adjust font size
            text = f"Class: {predicted_class_name}, Confidence: {confidence:.2f}"
            draw.text((10, 10), text, fill=(57, 255, 20), font=font)  # Adjust font color to black

            # Save the modified image
            image.save(image_dest_path)

        except IndexError:
            print(f"No matching class name found for prediction: {prediction}")

    # Save metrics to a text file
    with open(os.path.join(output_folder, 'metrics.txt'), 'w') as f:
        for metric, value in metrics.items():
            f.write(f"{metric}: {value}\n")

    # Save config file in the output folder
    with open(os.path.join(output_folder, 'config_test.yaml'), 'w') as f:
        yaml.dump(config, f)

    # Generate and save confusion matrix
    cm = confusion_matrix(all_targets, all_predictions)
    plt.figure(figsize=(8, 6))
    sns.heatmap(cm, annot=True, cmap="Blues", fmt="d")
    # Get the class names
    class_names = list(class_map_str_to_idx.keys())
    # Set tick labels and rotation
    plt.xticks(np.arange(len(class_names)) + 0.5, class_names, rotation=0)
    plt.yticks(np.arange(len(class_names)) + 0.5, class_names, rotation=0)
    plt.title('Confusion Matrix')
    plt.xlabel('Predicted Label')
    plt.ylabel('True Label')
    plt.tight_layout()
    plt.savefig(os.path.join(output_folder, 'confusion_matrix.png'))


    # Print where the results are stored
    print(f"\nResults are stored in: {output_folder}")

# Main function
def main(config_file):
    # Read configuration
    config = read_config(config_file)

    # Set up output folder
    output_folder = config['output_folder']
    if os.path.exists(output_folder):
        print("Output folder already exists. Exiting.")
        return

    # Set hyperparameters
    num_classes = config['num_classes']
    model_name = config['model_name']
    batch_size = config['batch_size']
    test_folder = config['test_folder']

    # Set up data transformations
    transform = transforms.Compose([
        transforms.Resize(config['image_size']),
        transforms.ToTensor(),
    ])

    # Load test dataset
    test_dataset = ImageFolder(test_folder, transform=transform)
    test_loader = DataLoader(test_dataset, batch_size=batch_size, shuffle=False)

    # Print the architecture selected
    print(f"Selected model architecture: {model_name}")

    # Determine if GPU or CPU is used
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print(f"Device in use: {device}")

    # Set up model
    model = setup_model(num_classes, model_name)

    # Load weights into the model
    weights_path = config['weights_path']
    print("Loading model weights...")
    load_weights(model, weights_path)
    print("Model weights loaded successfully.")

    # Test the model
    all_predictions, all_targets, test_loss, all_confidences = test_model(model, test_loader)

    # Calculate metrics
    accuracy = accuracy_score(all_targets, all_predictions)
    precision = precision_score(all_targets, all_predictions, average='weighted')
    recall = recall_score(all_targets, all_predictions, average='weighted')
    f1 = f1_score(all_targets, all_predictions, average='weighted')

    # Save results
    metrics = {'Accuracy': accuracy, 'Precision': precision, 'Recall': recall, 'F1 Score': f1}
    save_results(output_folder, all_predictions, all_targets, metrics, config, test_dataset, all_confidences)

# Run the code
if __name__ == "__main__":
    config_file = 'config_test.yaml'  # Change this to your config file path
    main(config_file)
