import yaml
import os
import sys
import shutil
import torch
import seaborn as sns
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, confusion_matrix
import matplotlib.pyplot as plt
import numpy as np
from torchvision.utils import make_grid, save_image
from main_esaim import (#change the from main to the file you used to train your federated dataset
    set_seeds,
    create_folder,
    get_model,
    get_criterion,
    get_optimizer,
    load_train_val_data,
    load_test_data,
    train_model,
    test_model,
    calculate_metrics,
    save_metrics,
)
from tqdm import tqdm

#### Set path to training dataset line 53 and Test at line 58!!!



# load config file
config_file_name = "config.yaml"
with open(config_file_name) as file:
    CONFIG = yaml.safe_load(file)

# specify output folder
save_dir = os.path.join("runs", CONFIG["output_folder"])

# check if folder already exists
if os.path.exists(save_dir):
    print("Output folder already exists. Exiting the program.")
    sys.exit()

# create folder
create_folder(save_dir)
print("Output folder created.")

# move copy of config file to output folder
shutil.copy(config_file_name, os.path.join(save_dir, config_file_name))
print("Config file copied to the output folder.")

# specify client dataset
cid = CONFIG["client_id"]

# set seeds
set_seeds(CONFIG["seed"])
print("Seeds set.")

# load pretrained model
model = get_model(CONFIG["model"], CONFIG["num_classes"])
print("Pretrained model loaded.")

# load data
print("Loading training and validation data...")
train_dataloader, val_dataloader = load_train_val_data(
    os.path.join("dataset/ESAIM", f"Client{int(cid)}"),
    CONFIG["batch_size"],
    CONFIG["img_size"],
)
print("Training and validation data loaded.")
print("Loading test data...")
test_dataloader = load_test_data(
    "dataset/ESAIM/global_test_set",
    CONFIG["batch_size"],
    CONFIG["img_size"],
)
print("Test data loaded.")

# save one batch of training data
imgs, _ = next(iter(train_dataloader))
train_batch0 = make_grid(imgs, nrow=4, scale_each=True)
save_image(train_batch0, os.path.join(save_dir, "train_batch0"), "png")
print("One batch of training data saved as image.")

# specify optimizer
optimizer = get_optimizer(model, CONFIG["optimizer"])
print("Optimizer specified.")

# specify loss function
criterion = get_criterion(CONFIG["loss_fcn"])
print("Loss function specified.")

# specify device
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
print(f"Device in use: {device}")

# train model
print("Training the model...")
metrics_dict = train_model(
    model,
    train_dataloader,
    val_dataloader,
    optimizer,
    CONFIG["epochs"],
    criterion,
    device,
    CONFIG["use_best_weights"],
    CONFIG["early_stopping"],
    save_dir,
)
print("Model training completed.")

# test model with best epoch weights
best_epoch_weights_path = os.path.join(save_dir, "best.pt")

if os.path.exists(best_epoch_weights_path):
    print("Loading best epoch weights...")
    model.load_state_dict(torch.jit.load(best_epoch_weights_path).state_dict())
    print("Best epoch weights loaded successfully.")
else:
    print("Best epoch weights not found. Using final weights for testing.")

# test model
print("Testing the model...")
test_acc, test_loss, confusion_matrix = test_model(
    model,
    test_dataloader,
    criterion,
    device,
    CONFIG["num_classes"],
)
print("Model testing completed.")


# calculate metrics
print("Calculating metrics...")
average_Precision, average_Recall, average_F1 = calculate_metrics(confusion_matrix)
print("Metrics calculated.")

# Determine the actual number of epochs completed
num_epochs_completed = len(metrics_dict.get("train_loss", []))

# Add test metrics to dict and pad the lists
metrics_dict["test_acc"] = [0] * (num_epochs_completed - 1) + [test_acc]
metrics_dict["test_loss"] = [0] * (num_epochs_completed - 1) + [test_loss]
metrics_dict["precision"] = [0] * (num_epochs_completed - 1) + [average_Precision]
metrics_dict["recall"] = [0] * (num_epochs_completed - 1) + [average_Recall]
metrics_dict["f1"] = [0] * (num_epochs_completed - 1) + [average_F1]

# write training and validation metrics to .txt file
print("Saving metrics to a file...")
save_metrics(metrics_dict, "epoch", os.path.join(save_dir, "results.txt"))
print("Metrics saved to results.txt in the output folder.")
print("\nResults are stored in:", save_dir)
